#include <vector>
#include <set>
#include <queue>
#include <algorithm>

#include "PathPlanning.h"
#include "../Core/Core.h"

namespace rm {

std::pair<std::vector<Node>, uint32_t> plan_one_path(const Grid& grid,
                                                     const std::set<Node>& revolve_area,
                                                     const Vehicle& vehicle,
                                                     const Node& from,
                                                     const Node& to,
                                                     uint32_t current_time,
                                                     bool is_first_map) {
    // 写此函数的时候地图已经确定了，为了赶工期，所以多处使用定值（逃
    constexpr uint32_t rows = 12;
    constexpr uint32_t cols = 12;
    // 目前使用的算法是 SPFA ，因为它能够方便地考虑方向因素
    auto queue = std::make_unique<std::queue<Node>>();
    auto in_queue = std::make_unique<grid_t<bool>>(rows, std::vector<bool>(cols, false));
    auto min_dist = std::make_unique<grid_t<uint32_t>>(rows, std::vector<uint32_t>(cols, UINT32_MAX));
    auto direction = std::make_unique<grid_t<uint32_t>>(rows, std::vector<uint32_t>(cols, UINT32_MAX));
    auto prev_node = std::make_unique<grid_t<Node>>(rows, std::vector<Node>(cols, {UINT32_MAX, UINT32_MAX}));

    // 不能从终点倒着走，由于有旋转门，倒着走和正着走的情况可能会是不一样的
    queue->push(from);
    // 不考虑起始点的方向
    (*min_dist)[from.row][from.col] = current_time - vehicle.speed_cor;
    while (!queue->empty()) {
        Node curr_node = queue->front();
        queue->pop();
        (*in_queue)[curr_node.row][curr_node.col] = false;
        uint32_t curr_dist = (*min_dist)[curr_node.row][curr_node.col];
        uint32_t curr_direction = (*direction)[curr_node.row][curr_node.col];
        for (const auto& next_node : {Node{curr_node.row + 1, curr_node.col},
                                      Node{curr_node.row - 1, curr_node.col},
                                      Node{curr_node.row, curr_node.col + 1},
                                      Node{curr_node.row, curr_node.col - 1}}) {
            if (!(next_node.row < rows && next_node.col < cols)
                || !grid.is_linked(curr_node, next_node)) continue;

            uint32_t next_dist = curr_dist;
            // 预估需要在旋转门前等待的时间
            // 缩小了可通过区间，避免可能的更长等待
            if (revolve_area.find(next_node) != revolve_area.end()) {
                uint32_t moded_time = next_dist % 60000;
                if (is_first_map) {
                    if (moded_time < 5000) next_dist += 5000 - moded_time;
                    else if (moded_time > 25000) next_dist += 65000 - moded_time;
                } else {
                    if (moded_time < 35000) next_dist += 35000 - moded_time;
                    else if (moded_time > 55000) next_dist += 95000 - moded_time;
                }
            }

            // 考虑方向因素
            uint32_t next_direction = get_direction(curr_node, next_node);
            if (next_direction == UP || next_direction == DOWN) {
                next_dist += vehicle.speed_col;
            } else {
                next_dist += vehicle.speed_row;
            }
            if (next_direction != curr_direction) next_dist += vehicle.speed_cor;

            if (next_dist < (*min_dist)[next_node.row][next_node.col]) {
                (*min_dist)[next_node.row][next_node.col] = next_dist;
                (*direction)[next_node.row][next_node.col] = next_direction;
                (*prev_node)[next_node.row][next_node.col] = curr_node;
                if (!(*in_queue)[next_node.row][next_node.col]) {
                    queue->push(next_node);
                    (*in_queue)[next_node.row][next_node.col] = true;
                }
            }
        }
    }

    // 两点不连通，输出空 std::vector 和 UINT32_MAX
    if ((*prev_node)[to.row][to.col] == Node{UINT32_MAX, UINT32_MAX}) {
        return {{}, UINT32_MAX};
    }

    std::vector<Node> path;
    Node temp_node = to;
    path.push_back(temp_node);
    do {
        temp_node = (*prev_node)[temp_node.row][temp_node.col];
        path.push_back(temp_node);
    } while (temp_node != from);
    std::reverse(path.begin(), path.end());
    return {path, (*min_dist)[to.row][to.col]};
}

std::pair<std::vector<Node>, uint32_t> plan_best_path(const Grid& grid1,
                                                      const Grid& grid2,
                                                      const std::set<Node>& revolve_area1,
                                                      const std::set<Node>& revolve_area2,
                                                      const Vehicle& vehicle,
                                                      const Node& from,
                                                      const Node& to,
                                                      uint32_t current_time) {
    // 此算法是短视的，若途中需要经过两个旋转门，实际上它是无法真正综合考虑两张图的
    // 它只会选择在其中一张图上最优的走法
    auto pair1 = plan_one_path(grid1, revolve_area1, vehicle, from, to, current_time, true);
    auto pair2 = plan_one_path(grid2, revolve_area2, vehicle, from, to, current_time, false);
    return pair1.second < pair2.second ? pair1 : pair2;
}

std::vector<Node> plan_best_order(const Grid& grid1,
                                  const Grid& grid2,
                                  const std::set<Node>& revolve_area1,
                                  const std::set<Node>& revolve_area2,
                                  const Vehicle& vehicle,
                                  const Node& from,
                                  std::vector<Node> dsts,
                                  uint32_t current_time) {
    std::sort(dsts.begin(), dsts.end());
    std::vector<Node> best_dst_order;
    uint32_t min_time = UINT32_MAX;
    constexpr uint32_t scratch_cost = 15000;
    do {
        Node last_node = from;
        std::vector<Node> path;
        uint32_t time = current_time;
        for (auto const& dst : dsts) {
            auto pair = plan_best_path(grid1, grid1, revolve_area1, revolve_area2,
                                       vehicle, last_node, dst, time);
            std::copy(pair.first.begin(), pair.first.end(), std::back_inserter(path));
            time += pair.second + scratch_cost;
            last_node = dst;
        }
        if (time < min_time) {
            min_time = time;
            best_dst_order = dsts;
        }
    } while (std::next_permutation(dsts.begin(), dsts.end()));
    return best_dst_order;
}

} // namespace
